from datetime import datetime
from typing import List
from langchain_core.messages import BaseMessage, FunctionMessage, ToolMessage
from pydantic import BaseModel, Json


class Function(BaseModel):
    arguments: Json
    name: str


class ToolCall(BaseModel):
    id: str
    function: Function
    type: str


class ToolCalls(BaseModel):
    calls: List[ToolCall]


def new_message_of_type(type: BaseMessage, *, content: str = "", **kwargs):
    return type(
        content=content,
        additional_kwargs={
            "timestamp": datetime.utcnow().timestamp(),
            **kwargs,
        },
    )


def is_function_or_tool_call(message: BaseMessage):
    return (
        message.additional_kwargs.get("tool_calls", None) is not None
        or message.additional_kwargs.get("function_call", None) is not None
    )


def is_tool_related_message(message: BaseMessage):
    return isinstance(
        message, (FunctionMessage, ToolMessage)
    ) or is_function_or_tool_call(message)



from __future__ import annotations

from datetime import datetime
from dataclasses import dataclass

import humanize
from rich.console import RenderResult, Console, ConsoleOptions
from rich.style import Style
from rich.color import Color
from rich.padding import Padding
from rich.text import Text
from textual import log, on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.message import Message
from textual.binding import Binding
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import Button, OptionList, Static
from textual.widgets.option_list import Option, Separator

from gptextual.textual_ui.widgets.footer import CommandFooter, Command, Field
from gptextual.runtime import ConversationManager


@dataclass
class ChatListItemRenderable:
    chat_id: str
    is_open: bool = False

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        chat = ConversationManager.get_conversation(self.chat_id)
        delta = datetime.utcnow().astimezone() - chat.update_time
        subtitle = humanize.naturaltime(delta)
        yield Padding(
            Text.assemble(
                (chat.short_preview, "" if not self.is_open else "b"),
                "\n\n",
                (f"{chat.model.name}@{chat.model.api_provider}", "i"),
                "\n",
                (subtitle, Style(color=Color.from_rgb(128, 128, 128))),
            ),
            pad=(0, 1),
            style="reverse" if self.is_open else "",
        )


class ChatListItem(Option):
    def __init__(self, chat_id: str, is_open: bool = False) -> None:
        """
        Args:
            chat: The chat associated with this option.
            is_open: True if this is the chat that's currently open.
        """
        super().__init__(ChatListItemRenderable(chat_id=chat_id, is_open=is_open))
        self.chat_id = chat_id
        self.is_open = is_open


class ChatList(Widget):
    BINDINGS = [
        Binding(
            "i",
            action="focus('chat-input')",
            description="Focus Input",
            key_display="i",
        ),
        Binding("r", "rename_conversation", "Rename Chat", key_display="r"),
        Binding("d", "delete_conversation", "Delete Chat", key_display="d"),
        Binding("e", "export", "Export To Markdown", key_display="e"),
    ]
    COMPONENT_CLASSES = {"app-title", "app-subtitle"}

    current_chat_id: var[str | None] = var(None)

    @dataclass
    class ChatOpened(Message):
        chat_id: str

    @dataclass
    class ChatDeleted(Message):
        chat_id: str

    def compose(self) -> ComposeResult:
        with Vertical(id="cl-header-container"):
            yield Static(
                Text(
                    "GPTextual",
                    style=self.get_component_rich_style("app-title"),
                )
            )
            yield Static(
                Text(
                    "LLMs in the Terminal",
                    style=self.get_component_rich_style("app-subtitle"),
                )
            )

        self.options = self.load_conversation_list_items()
        option_list = OptionList(
            *self.options,
            id="cl-option-list",
        )
        yield option_list

        with Horizontal(id="cl-button-container"):
            yield Button("New Chat", id="cl-new-chat-button")

    @on(OptionList.OptionSelected, "#cl-option-list")
    def post_chat_opened(self, event: OptionList.OptionSelected) -> None:
        assert isinstance(event.option, ChatListItem)
        self.post_message(ChatList.ChatOpened(chat_id=event.option.chat_id))
        self.current_chat_id = event.option.chat_id

    def watch_current_chat_id(self, old_chat_id: str, new_chat_id: str):
        if old_chat_id != new_chat_id:
            self.reload_and_refresh()

    def action_rename_conversation(self):
        if self.current_chat_id:
            footer: CommandFooter = self.app.query_one(CommandFooter)
            if footer.command:
                return

            conv = ConversationManager.get_conversation(self.current_chat_id)

            def rename(values):
                title: str = values[0]
                if title:
                    conv.title = title
                    conv.set_dirty()
                    conv.save()
                    self.reload_and_refresh()
                self.query_one("#cl-option-list").focus()

            fields = (
                Field(
                    "Title",
                    str,
                    has_spaces=True,
                    value=conv.title or conv.short_preview,
                ),
            )
            footer.command = Command("New Title", fields, rename)
            self.screen.set_focus(footer)

    def action_delete_conversation(self):
        if self.current_chat_id:
            footer: CommandFooter = self.app.query_one(CommandFooter)
            if footer.command:
                return

            def delete_chat(values):
                confirm: bool = values[0]
                if confirm:
                    ConversationManager.delete_conversation(self.current_chat_id)
                    self.post_message(
                        ChatList.ChatDeleted(chat_id=self.current_chat_id)
                    )
                    self.current_chat_id = None
                self.query_one(OptionList).focus()

            fields = (Field("yes/no", bool),)
            footer.command = Command("Delete this Chat?", fields, delete_chat)
            self.screen.set_focus(footer)

    def action_export(self):
        if self.current_chat_id:
            chat = ConversationManager.get_conversation(self.current_chat_id)
            path = chat.export_to_markdown()
            self.notify(
                f"Chat was saved as markdown at {path}",
                title="Export successful",
                timeout=10,
            )

    def on_focus(self) -> None:
        log.debug("Sidebar focused")
        self.query_one("#cl-option-list", OptionList).focus()

    def reload_and_refresh(self) -> None:
        """Reload the chats and refresh the widget. Can be used to
        update the ordering/previews/titles etc contained in the list.

        Args:
            new_highlighted: The index to highlight after refresh.
        """
        self.options = self.load_conversation_list_items()
        highlighted_idx = 0
        for idx, item in enumerate(self.options):
            if not isinstance(item, Separator) and item.chat_id == self.current_chat_id:
                highlighted_idx = idx // 2
        option_list = self.query_one(OptionList)
        option_list.clear_options()
        option_list.add_options(self.options)
        option_list.highlighted = highlighted_idx

    def load_conversation_list_items(self) -> list[ChatListItem]:
        chats = ConversationManager.all_conversations()
        ans = []
        for chat in chats:
            ans.append(ChatListItem(chat.id, is_open=self.current_chat_id == chat.id))
            ans.append(Separator())
        return ans



import glob
from pathlib import Path
import os
import threading
import logging
from collections import deque
from types import SimpleNamespace
from datetime import datetime
from dataclasses import dataclass, field
import json
from typing import List
import polars as pl

from langchain_core.messages import (
    AIMessageChunk,
    BaseMessage,
    AIMessage,
    HumanMessage,
    FunctionMessage,
    ToolMessage,
    SystemMessage,
)
from langchain_core.outputs import ChatGenerationChunk
from langchain_core.language_models import BaseLanguageModel


from shortuuid import ShortUUID

from gptextual.runtime.langchain.schema import (
    is_function_or_tool_call,
    is_tool_related_message,
)
from gptextual.runtime.function_calling import FunctionCallSupport
from gptextual.runtime.models import ModelRegistry, ChatModel
from gptextual.logging import logger


@dataclass
class StreamingMessage:
    message: BaseMessage = None
    additional_kwargs: dict = field(
        default_factory=lambda: {"timestamp": datetime.utcnow().timestamp()}
    )


def ensure_list(x):
    if not isinstance(x, list):
        return [x]
    return x


conversation_path = Path.home() / (".gptextual") / "conversations"
export_path = Path.home() / (".gptextual") / "exports"
MESSAGE_COLUMNS = SimpleNamespace(
    id="id", type="type", content="content", additionals="additional_kwargs"
)


@dataclass
class Conversation:
    id: str | None
    model: ChatModel
    title: str | None
    create_timestamp: float | None

    # In-memory conversations can be used for functions that spawn
    # side conversations with different LLMs. These should not be persisted.
    in_memory: bool = False
    messages: list[BaseMessage] = field(init=False, default_factory=list)

    def __post_init__(self):
        self._on_chunk_callback = None
        self._dirty = False
        self._messages_lock = threading.Lock()
        self.uuid_gen = ShortUUID()

    def __len__(self):
        return len(self.messages)

    def __iter__(self):
        return iter(self.messages)

    def set_dirty(self):
        self._dirty = True
        Conversation.search_df = None

    def append(self, message: BaseMessage | List[BaseMessage]):
        message = ensure_list(message)
        with self._messages_lock:
            for m in message:
                if "id" not in m.additional_kwargs:
                    m.additional_kwargs["id"] = self.uuid_gen.random(20)
                self.messages.append(m)
        self.set_dirty()

    @property
    def first_user_message(self) -> BaseMessage | None:
        for m in self.messages:
            if isinstance(m, HumanMessage):
                return m

    def last_message_of_types(self, *types):
        for m in reversed(self.messages):
            if isinstance(m, tuple(types)):
                return m
        return None

    @property
    def short_preview(self):
        if self.title:
            return self.title[:30]
        return Conversation.preview_from_messages(messages=self.messages)

    @property
    def displayable_messages(self) -> list[BaseMessage]:
        return [
            m
            for m in self.messages
            if not isinstance(
                m, (StreamingMessage, SystemMessage, ToolMessage, FunctionMessage)
            )
            and not is_function_or_tool_call(m)
        ]

    @property
    def create_time(self) -> datetime:
        return datetime.fromtimestamp(self.create_timestamp or 0).astimezone()

    @property
    def update_time(self) -> datetime:
        create_time = self.create_time.timestamp()
        return datetime.fromtimestamp(
            self.messages[-1].additional_kwargs.get("timestamp", create_time)
            if self.messages
            else create_time
        ).astimezone()

    @property
    def streaming_message(self):
        if not len(self):
            return None
        return (
            self.messages[-1]
            if isinstance(self.messages[-1], StreamingMessage)
            else None
        )

    @property
    def on_stream_chunk(self):
        return self._on_chunk_callback

    @on_stream_chunk.setter
    def on_stream_chunk(self, cb):
        self._on_chunk_callback = cb

    def has_messages_of_types(self, *types):
        return any([isinstance(m, tuple(types)) for m in self.messages])

    def progress(self, messages: BaseMessage | List[BaseMessage], *, autosave=True):
        return self._progress_llm(messages=messages, autosave=autosave)

    async def _progress_llm(
        self, messages: BaseMessage | List[BaseMessage], *, autosave=True
    ):
        try:
            if self.streaming_message:
                logger().info(
                    f"Conversation {self.title or self.id} is already streaming, but another message was sent to it: {messages}"
                )
                return

            messages = ensure_list(messages)

            for m in messages:
                if isinstance(m, (AIMessage, SystemMessage)):
                    raise ValueError(
                        "Conversation can only be progressed by human or function/tool messages"
                    )

            self.append(messages)
            response = StreamingMessage()
            self.append(response)
            chunks = 0
            async for chunk in self._stream_llm(
                self._messages_for_context_size(self.messages)[:-1]
            ):
                response.message = (
                    chunk if response.message is None else response.message + chunk
                )

                if self.on_stream_chunk:
                    await self.on_stream_chunk(response, chunks)
                chunks += 1
                yield response

            self.messages.pop()
            response = response.message
            if response:
                self.append(response)

            function_calling = FunctionCallSupport.forModelName(
                model_name=self.model.name, api_provider=self.model.api_provider
            )
            function_results = (
                await function_calling.execute_function_call(response)
                if function_calling
                else None
            )

            if function_results:
                async for resp in self._progress_llm(function_results, autosave=False):
                    response = resp
                    yield response
            else:
                response.additional_kwargs["timestamp"] = datetime.utcnow().timestamp()
                if self.on_stream_chunk:
                    await self.on_stream_chunk(response, chunks)
                yield response

            if autosave:
                self.save(in_background=True)

        except Exception as ex:
            logger().error(f"Error progressing the LLM conversation: {ex}")
            yield AIMessageChunk(content=f"There was conversation error, {ex}")

    async def _stream_llm(self, messages):
        try:
            function_kwargs = {}
            function_calling = FunctionCallSupport.forModelName(
                model_name=self.model.name, api_provider=self.model.api_provider
            )
            if function_calling:
                function_kwargs = function_calling.get_kwargs()

            if logger().getEffectiveLevel() <= logging.INFO:
                log_msg = messages[-3:] if len(messages) >= 3 else [*messages]
                log_msg.reverse()
                logger().info(
                    f"Calling model {self.model.name}@{self.model.api_provider}",
                    extra={
                        "last_3_messages": [
                            {
                                "type": m.type,
                                "content": m.content,
                                "kwargs": m.additional_kwargs,
                            }
                            for m in log_msg
                        ],
                        "functions": function_kwargs,
                    },
                )

            def to_message(chunk):
                if isinstance(chunk, str):
                    return AIMessageChunk(content=chunk)
                if isinstance(chunk, BaseMessage):
                    return chunk
                elif isinstance(chunk, ChatGenerationChunk):
                    return chunk.message

                raise ValueError(
                    f"Cannot get message from unknown chunk type {chunk.__class__}"
                )

            async for chunk in self.model.llm_model.astream(
                messages, **function_kwargs
            ):
                try:
                    yield to_message(chunk)
                except Exception as ex:
                    raise ex
        except Exception as ex:
            logger().error(f"Error during LLM streaming: {ex}")
            yield AIMessageChunk(
                content=f"There was an error streaming the LLM response, {ex}"
            )

    def _get_message_length(self, message: BaseMessage) -> int:
        if isinstance(message, StreamingMessage):
            return self.model.default_max_tokens
        else:
            model_name = self.model.name
            model = self.model.llm_model
            return Conversation.get_num_tokens_from_messages(
                messages=message, model_name=model_name, model=model
            )[0]

    def _messages_for_context_size(
        self, messages: list[BaseMessage]
    ) -> list[BaseMessage]:
        """Returns a list of messages that fit within the context window."""

        if not messages:
            return []

        context_messages = deque()

        system_message = messages[0] if isinstance(messages[0], SystemMessage) else None
        length_available = self.model.context_window
        if system_message:
            length_available -= self._get_message_length(system_message)
            if length_available < 0:
                raise ValueError("System message is too long for context window")

        for message in reversed(messages):
            message_length = self._get_message_length(message)
            if message_length <= length_available:
                context_messages.appendleft(message)
                length_available -= message_length
            else:
                break

        if context_messages and context_messages[0] is not system_message:
            context_messages.appendleft(system_message)

        return list(context_messages)

    def save(self, in_background=True):
        if self.in_memory:
            return

        def do_save():
            try:
                if not self._dirty:
                    return

                data = {
                    "id": self.id,
                    "model": self.model.name,
                    "api_provider": self.model.api_provider,
                    "title": self.title,
                    "create_timestamp": self.create_timestamp,
                }

                # Save the data to a JSON file
                with open(conversation_path / f"{self.id}.json", "w") as f:
                    json.dump(data, f)

                rows = {
                    MESSAGE_COLUMNS.id: [],
                    MESSAGE_COLUMNS.type: [],
                    MESSAGE_COLUMNS.content: [],
                    MESSAGE_COLUMNS.additionals: [],
                }

                # Filter out tool related messages
                with self._messages_lock:
                    messages = list(
                        filter(lambda m: not is_tool_related_message(m), self.messages)
                    )
                    self.messages = messages

                for m in messages:
                    rows[MESSAGE_COLUMNS.id].append(m.additional_kwargs["id"])
                    rows[MESSAGE_COLUMNS.type].append(m.type)
                    rows[MESSAGE_COLUMNS.content].append(m.content)
                    rows[MESSAGE_COLUMNS.additionals].append(
                        json.dumps(m.additional_kwargs)
                    )
                df = pl.DataFrame(rows)
                df.write_parquet(conversation_path / f"{self.id}.parquet")
                self._dirty = False
            except Exception as ex:
                logger().error(
                    f"There was an error saving the conversation {self.id}: {ex}"
                )

        if in_background:
            save_thread = threading.Thread(target=do_save)
            # Start the new thread
            save_thread.start()
        else:
            do_save()

    def export_to_markdown(self):
        # Create the markdown header
        create_time = datetime.fromtimestamp(self.create_timestamp).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
        header = f"# Conversation: {self.title}\n"
        header += f"Model: {self.model.name}\n\n"
        header += f"API Provider: {self.model.api_provider}"
        header += f"Created At: {create_time}\n"
        header += "\n---\n\n"  # Add a horizontal rule for styling

        def message_prefix(m):
            type = m.type
            type = type.lower()

            if type.startswith("human"):
                return "Human"
            elif type.startswith("ai"):
                return "AI"
            elif type.startswith("system"):
                return "System"
            elif type.startswith("tool"):
                return "Tool"
            elif type.startswith("func"):
                return "Function"

        # Create markdown entries for each message
        messages_md = ""
        for message in self.messages:
            prefix = f"**{message_prefix(message)}**\n\n"
            messages_md += f"{prefix}{message.content}\n\n"

        # Combine the header and messages
        markdown_content = header + messages_md

        # Write the markdown content to a file
        markdown_path = export_path / f"{self.model.name}_{create_time}.md"
        with open(markdown_path, "w", encoding="utf-8") as md_file:
            md_file.write(markdown_content)
        return markdown_path

    @classmethod
    def search(cls, query: str) -> pl.DataFrame:
        if cls.search_df is None:
            try:
                files = glob.glob(str(conversation_path / "*.parquet"))

                search_df = None
                for file in files:
                    filepath = Path(file)
                    conv_id = filepath.stem
                    header_file = conversation_path / f"{conv_id}.json"
                    header_data = None
                    with open(header_file, "r") as f:
                        header_data = json.load(f)

                    if (
                        header_data is None
                        or ModelRegistry.get_instance().model_from_name(
                            header_data["model"], header_data["api_provider"]
                        )
                        is None
                    ):
                        continue

                    df = (
                        pl.read_parquet(filepath)
                        .select(MESSAGE_COLUMNS.id, MESSAGE_COLUMNS.content)
                        .with_columns(
                            pl.lit(conv_id).alias("conv_id"),
                            pl.col(MESSAGE_COLUMNS.content)
                            .str.to_lowercase()
                            .alias("search_content"),
                        )
                    )
                    search_df = (
                        pl.concat([search_df, df]) if search_df is not None else df
                    )
            except Exception as ex:
                logger().error(f"Error loading conversation search data: {ex}")

            if search_df is not None:
                cls.search_df: pl.DataFrame = search_df
            else:
                cls.search_df = pl.DataFrame(
                    schema=[
                        (MESSAGE_COLUMNS.id, pl.Utf8),
                        (MESSAGE_COLUMNS.content, pl.Utf8),
                        ("conv_id", pl.Utf8),
                        ("search_content", pl.Utf8),
                    ]
                )

        res = (
            cls.search_df.filter(
                pl.col("search_content").str.contains(query.lower(), literal=True)
            )
            .select(MESSAGE_COLUMNS.id, "conv_id", MESSAGE_COLUMNS.content)
            .head(20)
        )
        return res

    @classmethod
    def load(cls, id: str):
        try:
            header_file = conversation_path / f"{id}.json"
            messages_file = conversation_path / f"{id}.parquet"

            if not os.path.exists(header_file) or not os.path.exists(messages_file):
                raise ValueError(
                    f"Conversation {id} cannot be loaded. Files do not exist."
                )

            # Load the data from the JSON file
            with open(header_file, "r") as f:
                data = json.load(f)

            # Replace the 'model' attribute with the actual model
            data["model"] = ModelRegistry.get_instance().model_from_name(
                data["model"], data["api_provider"]
            )

            if not data["model"]:
                return None
            del data["api_provider"]
            # Create a new Conversation object from the loaded data
            conv = cls(**data)
            df = pl.read_parquet(messages_file)
            for row in df.rows(named=True):
                type, content, kwargs = (
                    row[MESSAGE_COLUMNS.type],
                    row[MESSAGE_COLUMNS.content],
                    row[MESSAGE_COLUMNS.additionals],
                )
                kwargs = {"content": content, "additional_kwargs": json.loads(kwargs)}
                type = type.lower()
                message = None

                if type.startswith("human"):
                    message = HumanMessage(**kwargs)
                elif type.startswith("ai"):
                    message = AIMessage(**kwargs)
                elif type.startswith("system"):
                    message = SystemMessage(**kwargs)
                elif type.startswith("tool"):
                    message = ToolMessage(**kwargs)
                elif type.startswith("func"):
                    message = FunctionMessage(**kwargs)

                if message:
                    conv.messages.append(message)

            conv._dirty = False
            return conv
        except Exception as ex:
            logger().error(f"There was an error loading conversation {id}: {ex}")
            return None

    @staticmethod
    def preview_from_messages(messages: list) -> str:
        first_user_message = [m for m in messages if isinstance(m, HumanMessage)]
        first_user_message = first_user_message[0] if len(first_user_message) else None
        if not first_user_message:
            return "Empty chat..."
        first_content = first_user_message.content or ""
        return first_content[:30] + "..."

    @classmethod
    def create_new(
        cls,
        *,
        model: ChatModel,
        title: str = None,
        system_message: str = None,
        in_memory: bool = False,
    ):
        if model is None:
            raise ValueError("No model provided for conversation")

        _system_message = f"""
The current UTC datetime is {datetime.utcnow()}.
Do not escape markdown as codeblocks using '`' chars.Messages will be rendered as markdown by default!
"""
        if system_message is not None:
            _system_message += f"\n{system_message}"
        conv = cls(
            id=ShortUUID().random(20),
            model=model,
            title=title,
            create_timestamp=datetime.utcnow().timestamp(),
            in_memory=in_memory,
        )
        conv.append(SystemMessage(content=_system_message))
        return conv

    @staticmethod
    def get_num_tokens_from_messages(
        *,
        messages: BaseMessage | list[BaseMessage],
        model_name: str,
        model: BaseLanguageModel,
    ) -> list[int]:
        messages = ensure_list(messages)

        if not messages:
            return 0

        try:
            return [model.get_num_tokens_from_messages(messages=[m]) for m in messages]
        except Exception as ex:
            logger().error(
                f"Error calculating number of token messages for model {model_name}: {ex}. Falling back on estimate."
            )
            # Cannot get exact token count...do very(!) rough estimation.
            return [int(len(m.content) / 3.5) for m in messages]


Conversation.search_df = None



from datetime import datetime, timezone
import asyncio


def format_timestamp(timestamp: float) -> str:
    """Convert a Unix timestamp into a string in human-readable format.

    Args:
        timestamp: The Unix timestamp.

    Returns:
        The string timestamp in the format "%Y-%m-%d %H:%M:%S".
    """
    utc_dt = datetime.fromtimestamp(timestamp, timezone.utc)
    local_dt = utc_dt.astimezone()
    return local_dt.strftime("%Y-%m-%d %H:%M:%S")


def single_css_class(name, prop, val):
    return f"""
    .{name} {'{'}
     {prop}: {val};   
    {'}'}
  """


class AsyncCache:
    def __init__(self):
        self.cache = {}
        self.in_progress = {}

    def __call__(self, func):
        async def wrapper(*args, **kwargs):
            key = str(args) + str(kwargs)
            if key not in self.cache:
                if key not in self.in_progress:
                    self.in_progress[key] = asyncio.create_task(func(*args, **kwargs))
                    self.cache[key] = await self.in_progress[key]
                    del self.in_progress[key]
                else:
                    await self.in_progress[key]
                    # Now the result is in the cache
            return self.cache[key]

        return wrapper



from dataclasses import dataclass
from typing import Optional

from langchain_core.language_models import BaseLanguageModel

from gptextual.config.app_config import APIProviderConfig

from gptextual.config import AppConfig, APIProvider


@dataclass
class ChatModel:
    name: str
    api_provider: str
    context_window: int = 4097
    _model: BaseLanguageModel = None

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, o: object) -> bool:
        return isinstance(o, ChatModel) and o.name == self.name

    @property
    def default_max_tokens(self):
        return int(min(self.context_window * 0.1, 2000))

    @property
    def llm_model(self):
        if not self._model:
            config = AppConfig.get_instance()
            api_provider_config: APIProviderConfig = getattr(
                config.api_config, self.api_provider
            )
            if api_provider_config is None:
                raise ValueError(
                    f"Configuration Error: There is no app configuration for API provider {self.api_provider}"
                )

            kwargs = {"temperature": 0.6}
            if self.api_provider == APIProvider.GOOGLE:
                kwargs["max_output_tokens"] = self.default_max_tokens
            else:
                kwargs["max_tokens"] = self.default_max_tokens

            self._model = api_provider_config.create_model_instance(self.name, **kwargs)

        return self._model


class ModelRegistry:
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = ModelRegistry()
        return cls._instance

    def register_model(self, *, api_provider: str, model: ChatModel):
        self.models[api_provider][model.name] = model

    @classmethod
    def model_from_name(cls, name: str, api_provider: APIProvider):
        inst = cls.get_instance()
        if api_provider in inst.models:
            return inst.models[api_provider].get(name, None)

        return None

    def initialize_models_for_provider(self, api_provider: str):
        app_config = AppConfig.get_instance()
        api_config = getattr(app_config.api_config, api_provider.replace("-", "_"))
        if api_config:
            models: dict = api_config.models
            self.models = {
                **self.models,
                api_provider: {
                    **{
                        name: ChatModel(
                            name=name,
                            api_provider=api_provider,
                            context_window=conf.context_window,
                        )
                        for name, conf in models.items()
                    },
                },
            }

    def __init__(self) -> None:
        self.models = {}
        for api_provider in APIProvider:
            self.initialize_models_for_provider(api_provider.value)

    def __len__(self):
        n_models = 0
        for api_provider in self.models:
            n_models += len(self.models[api_provider])

        return n_models

    def all_models(self):
        for api_provider in self.models:
            for model in self.models[api_provider].values():
                yield model

    def models_for_provider(self, api_provider: str):
        if api_provider in self.models:
            for model in self.models[api_provider].values():
                yield model

    @property
    def default_model(self) -> ChatModel:
        ans = ModelRegistry.model_from_name("gpt-3.5-turbo", APIProvider.OPEN_AI)
        if ans is None:
            for m in self.all_models():
                return m
        return ans


@dataclass
class AppContext:
    """
    App Context
    """

    chat_message: Optional[str] = None
    model_name: str = None
    api_provider: str = None

    @property
    def current_model(self) -> ChatModel:
        registry = ModelRegistry.get_instance()
        if self.model_name is None:
            return registry.default_model
        model = ModelRegistry.model_from_name(self.model_name, self.api_provider)
        if model is None:
            raise ValueError(
                f"Current Model {self.model_name}@{self.api_provider} not found"
            )
        return model



from typing import Any
from dataclasses import dataclass
import inspect

from textual.widgets import Footer, Input, Label
from textual.containers import Horizontal, Vertical
from rich.console import RenderableType
from textual.reactive import reactive

from gptextual.utils import single_css_class


@dataclass
class Field:
    name: str
    type: Any
    has_spaces: bool = False
    value: str = ""

    def __str__(self):
        return f"<{self.name}: {self.type.__name__}>"


@dataclass
class Command:
    name: str
    fields: tuple
    on_submit: object


newline_char = "\n"


class CommandFooter(Footer):
    height_pct = [
        single_css_class(f"h-pct-{val}", "height", f"{val}%")
        for val in range(0, 110, 10)
    ]
    height = [single_css_class(f"h-{val}", "height", f"{val}") for val in range(0, 11)]
    width_pct = [
        single_css_class(f"w-pct-{val}", "width", f"{val}%")
        for val in range(0, 110, 10)
    ]
    OTHER_CSS = """
      #command_input {
        min-height: 4;
      }
      .align-center {
        align: center middle;
      } 

      .ml-2 {
        margin-left: 2;
      }
    """

    DEFAULT_CSS = f"""
    {newline_char.join(height_pct)}
    {newline_char.join(height)} 
    {newline_char.join(width_pct)}
    {OTHER_CSS}
    """

    command: Command = reactive(None)

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        self.height_class = "h-4"

    def render(self) -> RenderableType:
        if self.command is not None:
            return ""

        return super().render()

    def watch_command(self, command: Command):
        self.update(command)

    @property
    def value(self):
        if self.command:
            ans = [f.value for f in self.command.fields if f.value]
            return " ".join(ans)

        return None

    @property
    def placeholder(self):
        if self.command:
            ans = [str(f) for f in self.command.fields]
            return " ".join(ans)
        return None

    def _extract_values(self, *args):
        if len(args) == 0 or args[0] is None:
            return None

        values = []
        for i in range(-1, -(len(self.command.fields) + 1), -1):
            field: Field = self.command.fields[i]
            if field.has_spaces:
                if i == -1:
                    values.append(" ".join(args))
                else:
                    values.append(" ".join(args[: i + 1]))
                break
            else:
                if field.type == bool:
                    values.append(args[i].lower() in ("y", "yes", "true"))
                else:
                    values.append(field.type(args[i]))

        return tuple(reversed(values))

    def update(self, command: Command):
        if command is None:
            try:
                container = self.query_one("#command_footer")
                container.remove()
            except Exception:
                pass
            self.remove_class(self.height_class)
            return

        input_field = Input(value=self.value, placeholder="", classes="w-pct-100 h-3")

        children = []

        children.append(
            Horizontal(
                Label(command.name, classes="align-center"),
                Vertical(
                    input_field,
                    Label(self.placeholder, classes="w-pct-100 h-1 ml-2"),
                    classes="w-pct-100 h-pct-100",
                ),
                classes="align-center h-pct-20",
                id="command_input",
            ),
        )

        self.mount(
            Vertical(
                *children,
                id="command_footer",
                classes="align-center h-pct-100 w-pct-100",
            )
        )
        self.add_class(self.height_class)
        self.screen.set_focus(input_field)

    def _on_key(self, event) -> None:
        if event.key == "escape":
            self.command = None

    async def on_input_submitted(self, message: Input.Submitted):
        if self.command is None:
            return
        try:
            callback = self.command.on_submit
            tokens = message.value.split(" ")
            values = self._extract_values(*tokens)
            ans = callback(values)
            if inspect.iscoroutine(ans):
                await ans
        except Exception:
            pass

        self.command = None



from __future__ import annotations

import os
from enum import Enum
import yaml
from pathlib import Path
from typing import Optional, Dict
import json

from langchain_core.language_models import BaseLanguageModel
from pydantic import BaseModel

try:
    from langchain_openai import ChatOpenAI as LangChainChatOpenAI
except ImportError:
    LangChainChatOpenAI = None

try:
    from langchain_google_genai import ChatGoogleGenerativeAI
except ImportError:
    ChatGoogleGenerativeAI = None

try:
    from gen_ai_hub.proxy.langchain.init_models import (
        init_llm as sap_genai_create_langchain_llm,
    )
    from gen_ai_hub.proxy.core.proxy_clients import set_proxy_version
except ImportError:
    sap_genai_create_langchain_llm = None
    set_proxy_version = None

from gptextual.logging import logger

SIZE_4K = 4096
SIZE_8K = 8192
SIZE_16K = 16385
SIZE_32K = 32768
SIZE_128K = 128000


class APIProvider(str, Enum):
    SAP_GEN_AI = "gen-ai-hub"
    OPEN_AI = "openai"
    GOOGLE = "google"


class TextualConfig(BaseModel):
    refresh_no_stream_chunks: int = 5
    theme: Optional[str] = "dark"


class ModelConfig(BaseModel):
    context_window: Optional[int] = SIZE_4K


class APIProviderConfig(BaseModel):
    function_call_support: Optional[Dict[str, str]] = {}
    models: Optional[Dict[str, ModelConfig | None]] = {}

    def create_config_file(self):
        ...

    def _write_config_file(self, *, folder: Path, filename: str, payload: str):
        os.makedirs(folder, exist_ok=True)
        with open(folder / filename, "w") as f:
            f.write(payload)

    def create_model_instance(self, model_name: str, **kwargs) -> BaseLanguageModel:
        return None


class GenAIHubConfig(APIProviderConfig):
    client_id: str
    client_secret: str
    auth_url: str
    api_base: str
    resource_group: str

    def create_config_file(self):
        folder = Path.home() / ".aicore"
        filename = "config.json"
        payload = {
            "AICORE_CLIENT_ID": self.client_id,
            "AICORE_CLIENT_SECRET": self.client_secret,
            "AICORE_AUTH_URL": self.auth_url,
            "AICORE_BASE_URL": self.api_base,
            "AICORE_RESOURCE_GROUP": self.resource_group,
        }
        self._write_config_file(
            folder=folder, filename=filename, payload=json.dumps(payload)
        )

    def create_model_instance(self, model_name: str, **kwargs) -> BaseLanguageModel:
        if set_proxy_version:
            set_proxy_version(APIProvider.SAP_GEN_AI.value)
        return (
            sap_genai_create_langchain_llm(model_name, **kwargs)
            if sap_genai_create_langchain_llm
            else None
        )


class OpenAIConfig(APIProviderConfig):
    api_key: str
    models: Optional[Dict[str, ModelConfig | None]] = {
        "gpt-4-0125-preview": ModelConfig(context_window=SIZE_128K),
        "gpt-4-turbo-preview": ModelConfig(context_window=SIZE_128K),
        "gpt-4": ModelConfig(context_window=SIZE_8K),
        "gpt-3.5-turbo-0125": ModelConfig(context_window=SIZE_16K),
        "gpt-3.5-turbo": ModelConfig(context_window=SIZE_4K),
    }

    def create_model_instance(self, model_name: str, **kwargs) -> BaseLanguageModel:
        return (
            LangChainChatOpenAI(model=model_name, openai_api_key=self.api_key, **kwargs)
            if LangChainChatOpenAI
            else None
        )


class GoogleConfig(APIProviderConfig):
    api_key: str
    models: Optional[Dict[str, ModelConfig | None]] = {
        "gemini-pro": ModelConfig(context_window=30720)
    }

    def create_model_instance(self, model_name: str, **kwargs) -> BaseLanguageModel:
        return (
            ChatGoogleGenerativeAI(
                model=model_name,
                google_api_key=self.api_key,
                convert_system_message_to_human=True,
                **kwargs,
            )
            if ChatGoogleGenerativeAI
            else None
        )


class APIConfig(BaseModel):
    gen_ai_hub: Optional[GenAIHubConfig] = None
    openai: Optional[OpenAIConfig] = None
    google: Optional[GoogleConfig] = None

    def _create_config_files(self):
        if self.gen_ai_hub:
            self.gen_ai_hub.create_config_file()


_instance: AppConfig = None


class AppConfig(BaseModel):
    functions: Optional[Dict[str, Optional[Dict[str, str]]]] = {}
    textual: Optional[TextualConfig] = TextualConfig()
    api_config: Optional[APIConfig] = APIConfig()
    log_level: Optional[str] = "INFO"

    @staticmethod
    def get_instance() -> AppConfig:
        global _instance
        if _instance is None:
            _instance = AppConfig.load_config()
            if _instance.api_config:
                _instance.api_config._create_config_files()
        return _instance

    @staticmethod
    def load_config() -> AppConfig:
        config_path = Path.home() / ".gptextual" / "config.yml"
        if config_path.exists():
            config_data = None
            with open(config_path, "r", encoding="utf-8") as f:
                config_data = yaml.safe_load(f)
            return AppConfig(**config_data) if config_data else AppConfig()
        else:
            # If the file is not found, return a default AppConfig instance
            logger().info(
                f"Configuration file not found at {config_path}. Using default settings."
            )
            return AppConfig()



from textual.app import ComposeResult
from textual.screen import Screen
from textual.binding import Binding
from textual.widget import Widget


from gptextual.textual_ui.widgets.search import Search


class SearchScreen(Screen):
    BINDINGS = [Binding("escape", "close", "Close", show=False)]

    DEFAULT_CSS = """
    SearchScreen {
      background: $panel 90%;
      content-align: center top;
      align: center top;
      width: 100%;
      height: 100%;
    }
    
    SearchScreen Search {
      width: 60;
      margin-top: 15;
      align: center middle;
      content-align: center middle;
      background: black 0%;
    }
    
  """

    def __init__(self, parent: Widget = None, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        self.search = Search(parent)

    def action_close(self):
        self.app.pop_screen()

    def on_mount(self):
        search = self.query_one(Search)
        search.input_field.focus()

    def compose(self) -> ComposeResult:
        yield self.search



from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.message import Message
from textual.widget import Widget
from textual.widgets import Static, Select

from gptextual.runtime import ChatModel, ModelRegistry


class ModelSelect(Widget, can_focus=True):
    BINDINGS = [
        Binding(
            key="ctrl+s",
            action="focus('cl-option-list')",
            description="Focus List",
            key_display="^s",
        ),
    ]

    class ModelSelected(Message):
        def __init__(self, model: ChatModel):
            super().__init__()
            self.model = model

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        registry = ModelRegistry.get_instance()
        default_model: ChatModel = registry.default_model
        if len(registry):
            self.provider_select = Select(
                id="provider-select",
                options=sorted(
                    [(p, p) for p in registry.models if len(registry.models[p])],
                    key=lambda x: x[0],
                ),
                allow_blank=False if default_model else True,
                value=default_model.api_provider if default_model else None,
            )
            self.model_select = Select(
                id="model-select",
                options=sorted(
                    [
                        (model.name, model)
                        for model in registry.models_for_provider(
                            default_model.api_provider
                        )
                    ]
                    if default_model
                    else [],
                    key=lambda x: x[0],
                ),
                allow_blank=False if default_model else True,
                value=default_model,
            )
        else:
            self.provider_select = Static(
                "There are no models configured in `~/.gptextual/config.yml`. Check the README on how to configure API providers and models."
            )
            self.model_select = None

    def _update_model_select(self, api_provider: str):
        registry = ModelRegistry.get_instance()
        options = list(
            sorted(
                [
                    (model.name, model)
                    for model in registry.models_for_provider(api_provider)
                ],
                key=lambda x: x[0],
            )
        )
        self.model_select.set_options(options)
        self.model_select.expanded = True
        self.model_select.refresh(layout=True)
        self.post_message(self.ModelSelected(model=self.model_select.value))

    @on(Select.Changed, "#provider-select")
    def provider_selected(self, event: Select.Changed) -> None:
        self._update_model_select(event.value)

    @on(Select.Changed, "#model-select")
    def model_selected(self, event: Select.Changed) -> None:
        self.post_message(self.ModelSelected(model=event.value))

    def compose(self) -> ComposeResult:
        with Vertical(id="chat-options-container"):
            if self.model_select:
                yield Static(
                    "Select an API provider and model. Then just start chatting to create a new conversation.",
                    id="model-select-label",
                )
                yield self.provider_select
                yield self.model_select
            else:
                yield self.provider_select



import threading

from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import Header, Input, OptionList
from textual.widgets.option_list import Option, Separator
from textual.binding import Binding, BindingType
from textual.containers import Vertical
from textual.message import Message
from textual import on
from textual.app import App

from gptextual.runtime import Conversation


class Search(Vertical):
    class SearchResultSelected(Message):
        def __init__(self, conversation_id: str, message_id: str) -> None:
            super().__init__()
            self.conversation_id = conversation_id
            self.message_id = message_id

    BINDINGS: list[BindingType] = [
        Binding("down", "cursor_down", "Down", show=False),
    ]

    DEFAULT_CSS = """

    .Search-expanded {
      height: 30;
    }

    Search {
      height: 3;
    }

    Search OptionList {
        border: tall $background;
        background: $panel;
        color: $text;
        padding: 0 1;

        width: 1fr;
        display: block;
        height: auto;
        max-height: 15;
        overlay: screen;
        constrain: y;
    }

    Search Input {
      width: 100%;
    }

  """

    def __init__(self, parent: Widget, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.timer = None
        self.parent_widget = parent

    @on(Input.Changed)
    def input_changed(self, event: Input.Changed):
        query = event.value
        if self.timer is not None:
            self.timer.cancel()

        # Create a new timer that calls the search method after a delay
        self.timer = threading.Timer(0.5, self.search, [query])
        self.timer.start()

    @on(OptionList.OptionSelected)
    def on_option_selected(self, event: OptionList.OptionSelected):
        conv_id, message_id = event.option_id
        widget = self.parent_widget if self.parent_widget else self
        widget.post_message(
            Search.SearchResultSelected(conversation_id=conv_id, message_id=message_id)
        )

    def set_options(self, options) -> None:
        select_options: list[Option] = [
            Option(prompt=prompt, id=value) for prompt, value in options
        ]

        option_list = self.query_one(OptionList)
        option_list.clear_options()
        for option in select_options:
            option_list.add_option(option)
            option_list.add_option(Separator())

    def action_cursor_down(self):
        input = self.query_one(Input)
        option_list = self.query_one(OptionList)
        if input.has_focus and option_list.option_count > 0:
            option_list.focus()
            option_list.highlighted = 0

    def break_into_lines(self, text, max_line_length):
        words = text.split(" ")
        lines = []
        current_line = []

        for word in words:
            if len(current_line) + len(word) + 1 <= max_line_length:
                current_line.append(word + " ")
            else:
                lines.append("".join(current_line).rstrip())
                current_line = [word + " "]

        # Add the last line if it's not empty
        if current_line:
            lines.append("".join(current_line).rstrip())

        return "\n".join(lines)

    def get_preview(self, content: str, query: str, max_line_length: 20):
        window_size = max(30, len(query))
        match_start_index = content.lower().find(query.lower())
        start_index = max(0, match_start_index - window_size)
        end_index = min(len(content), match_start_index + window_size + len(query))

        preview = f'{"..." if start_index > 0 else ""}{content[start_index:end_index]}{"..." if end_index < len(content) else ""}'

        return self.break_into_lines(preview, max_line_length=max_line_length)

    def search(self, query: str):
        results = []
        if query:
            df = Conversation.search(query)
            rows = df.rows(named=True)
            for row in rows:
                conv_id, message_id, content = row["conv_id"], row["id"], row["content"]
                results.append(
                    (
                        self.get_preview(content, query, max_line_length=30),
                        (conv_id, message_id),
                    )
                )

        options: OptionList = self.query_one(OptionList)
        self.set_options(results)
        expanded = len(results) > 0
        options.visible = expanded

        classname = "Search-expanded"
        if expanded:
            self.add_class(classname)
        else:
            self.remove_class(classname)

    @property
    def input_field(self):
        return self.query_one(Input)

    def compose(self) -> ComposeResult:
        options = OptionList(None)
        options.visible = False
        with Vertical():
            yield Input(placeholder="Search messages...")
            yield options


if __name__ == "__main__":

    class SelectApp(App):
        DEFAULT_CSS = """
    Screen {
      align: center top;
      background: blue;
    }

    Screen {
      background: blue;
    }
    Search {
      width: 60;
      margin: 2;
      background: $panel 0%;
    }
        """

        def compose(self) -> ComposeResult:
            yield Header()
            yield Search()

    app = SelectApp()
    app.run()



import logging
import logging.config
from logging.config import DictConfigurator as _dictConfig
import datetime as dt
import json
import queue
from pathlib import Path
import atexit
import os

import yaml


_logger = None


LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


class JSONFormatter(logging.Formatter):
    def __init__(
        self,
        *,
        fmt_keys: dict[str, str] | None = None,
    ):
        super().__init__()
        self.fmt_keys = fmt_keys if fmt_keys is not None else {}

    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord):
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.timezone.utc
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            always_fields["stack_info"] = self.formatStack(record.stack_info)

        message = {
            key: msg_val
            if (msg_val := always_fields.pop(val, None)) is not None
            else getattr(record, val)
            for key, val in self.fmt_keys.items()
        }
        message.update(always_fields)

        for key, val in record.__dict__.items():
            if key not in LOG_RECORD_BUILTIN_ATTRS:
                message[key] = val

        return message


class DictConfigurator(_dictConfig):
    def configure(self) -> None:
        super().configure()
        return self.config


class QueueHandler(logging.handlers.QueueHandler):
    def __init__(self, handlers):
        # Initialize the queue
        self.log_queue = queue.Queue(-1)  # No limit on queue size
        super().__init__(self.log_queue)

        # Initialize the handlers
        self.handlers = handlers

    def _initialize_listener(self, handlers):
        # Create and start the listener
        handlers = [handlers[h] for h in self.handlers]
        self.listener = logging.handlers.QueueListener(
            self.log_queue, *handlers, respect_handler_level=True
        )
        self.listener.start()

        # Register the stop method to be called at exit
        atexit.register(self.stop_listener)

    def stop_listener(self):
        """Stop the QueueListener."""
        self.listener.stop()


log_path = Path.home() / ".gptextual" / "logging"


def default_config(log_level: str = "INFO"):
    return f"""
  version: 1
  disable_existing_loggers: False
  formatters:
    simple:
      format: '%(asctime)s - %(levelname)s - %(pathname)s(%(lineno)d)::%(funcName)s - %(message)s'
    json:
      (): gptextual.logging.JSONFormatter
      fmt_keys:
        level: levelname
        message: message
        timestamp: timestamp
        logger: name
        module: module
        function: funcName
        line: lineno
        thread_name: threadName

  handlers:
    file_handler:
      class: logging.handlers.RotatingFileHandler
      level: DEBUG
      formatter: json 
      filename: {log_path / 'gptextual.jsonl'}
      maxBytes: 3485760 # 3MB
      backupCount: 1
      encoding: utf8
    queue_handler:
      (): gptextual.logging.QueueHandler
      handlers:
        - file_handler

  loggers:
    root:
      level: {log_level}  
      handlers:
        - queue_handler
  
  
  """


def setup_logging(log_level: str = "INFO"):
    global _logger
    os.makedirs(name=log_path, exist_ok=True)
    config = yaml.safe_load(default_config(log_level))
    configurator = DictConfigurator(config)
    config = configurator.configure()
    queue_handler = config["handlers"]["queue_handler"]
    queue_handler._initialize_listener(config["handlers"])
    _logger = logging.getLogger("gptextual")


def logger():
    return _logger



from __future__ import annotations

import json
from langchain.schema import BaseMessage
from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll, Vertical, Horizontal
from textual.screen import ModalScreen
from textual.widgets import Static, Tabs, ContentSwitcher, Tab

from gptextual.utils import format_timestamp
from gptextual.textual_ui.widgets.token_analysis import TokenAnalysis


class MessageInfo(ModalScreen):
    BINDINGS = [Binding("escape", "app.pop_screen", "Close Modal")]

    def __init__(
        self,
        message: BaseMessage,
        model_name: str,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
        )
        self.message = message
        self.model_name = model_name
        self.token_analysis = TokenAnalysis(
            text=self.message.content or "", model_name=model_name, id="tokens"
        )

    def compose(self) -> ComposeResult:
        with Vertical(id="outermost-container"):
            with Horizontal(id="message-info-header"):
                yield Tabs(
                    Tab("Markdown", id="markdown-content"),
                    Tab("Tokens", id="tokens"),
                    Tab("Additional Data", id="metadata"),
                )

            with VerticalScroll(id="inner-container"):
                with ContentSwitcher(initial="markdown-content"):
                    yield Static(self.message.content or "", id="markdown-content")
                    yield self.token_analysis
                    yield Static(
                        json.dumps(self.message.additional_kwargs, indent=4),
                        id="metadata",
                    )

            with Horizontal(id="message-info-footer"):
                if self.model_name:
                    token_count = len(self.token_analysis.tokens)

                timestamp = self.message.additional_kwargs.get("timestamp", 0)
                timestamp_string = format_timestamp(timestamp)
                yield Static(f"Message sent at {timestamp_string}", id="timestamp")
                yield Static(f"{token_count} tokens", id="token-count")

    @on(Tabs.TabActivated)
    def tab_activated(self, event: Tabs.TabActivated) -> None:
        self.query_one(ContentSwitcher).current = event.tab.id



import os
from typing import Optional

from textual.app import App

from toolong.watcher import get_watcher

from gptextual.logging import setup_logging, log_path
from gptextual.config import AppConfig
from gptextual.runtime.function_calling.function_call_support import (
    load_function_entry_points,
)
from gptextual.runtime.models import AppContext
from gptextual.runtime.conv_manager import ConversationManager
from gptextual.runtime.conversation import conversation_path, export_path

from gptextual.textual_ui.screens import ChatScreenDark, ChatScreenLight
from gptextual.textual_ui.widgets.footer import CommandFooter, Command, Field


class GPTextual(App):
    def __init__(self, context: Optional[AppContext] = None) -> None:
        super().__init__()
        app_config = AppConfig.get_instance()
        setup_logging(app_config.log_level)
        load_function_entry_points()
        self.app_context = context or AppContext()
        os.makedirs(conversation_path, exist_ok=True)
        os.makedirs(export_path, exist_ok=True)
        ConversationManager.load_conversations_from_storage()

        # attributes to integrate TooLong log file app
        self.file_paths = [str(log_path / "gptextual.jsonl")]
        self.merge = True
        self.save_merge = False
        self.watcher = None

    def start_log_watcher(self):
        if self.watcher:
            return

        self.watcher = get_watcher()
        self.watcher.start()

    def stop_log_watcher(self):
        if self.watcher:
            self.watcher.close()
            self.watcher = None

    def action_quit(self):
        footer: CommandFooter = self.query_one(CommandFooter)
        if footer.command:
            return

        def exit(values):
            confirm: bool = values[0]
            if confirm:
                self.exit()

        fields = (Field("yes/no", bool),)
        footer.command = Command("Quit Application?", fields, exit)
        self.screen.set_focus(footer)

    def on_mount(self) -> None:
        config = AppConfig.get_instance().textual
        self.push_screen(
            ChatScreenLight() if config.theme == "light" else ChatScreenDark()
        )

    def on_unmount(self) -> None:
        self.stop_log_watcher()


def run():
    app = GPTextual()
    app.run()


if __name__ == "__main__":
    run()



from __future__ import annotations

import pkg_resources
from asyncio import iscoroutine
from enum import Enum
from functools import cache
from typing import List
from langchain_core.utils.function_calling import (
    convert_to_openai_function,
    convert_to_openai_tool,
)
from langchain_core.messages import AIMessage, FunctionMessage, ToolMessage

from gptextual.config.app_config import APIProviderConfig, AppConfig
from gptextual.logging import logger
from gptextual.runtime.langchain.schema import Function, ToolCalls

# Central repository for registered functions
_FUNCTIONS_BY_NAME = {}


class FunctionCallSupport(str, Enum):
    OPENAI_TOOL = "openai_tool"  # new parallel function calling API
    OPENAI_FUNCTION = "openai_function"  # old singular function call API

    @staticmethod
    def fromStr(name: str) -> FunctionCallSupport:
        if name == FunctionCallSupport.OPENAI_TOOL:
            return FunctionCallSupport.OPENAI_TOOL
        elif name == FunctionCallSupport.OPENAI_FUNCTION:
            return FunctionCallSupport.OPENAI_FUNCTION

        return None

    @staticmethod
    @cache
    def forModelName(*, model_name: str, api_provider: str):
        app_config = AppConfig.get_instance()
        function_support = _DEFAULT_FUNCTION_CALL_SUPPORT.get(api_provider, {})
        api_provider_config: APIProviderConfig = getattr(
            app_config.api_config, api_provider
        )
        if api_provider_config and api_provider_config.function_call_support:
            function_support = {
                **function_support,
                **api_provider_config.function_call_support,
            }

        # 1. Check if there are any tools registered for the current proxy type
        if function_support:
            support_by_model = {
                model_name.strip(): support
                for key, support in function_support.items()
                for model_name in key.split(",")
            }

            # 2. Check for use of wildcard (all models)
            if model_name not in support_by_model:
                model_name = "*"
            if model_name in support_by_model:
                return FunctionCallSupport.fromStr(support_by_model[model_name])

        return None

    async def execute_function_call(self, message: AIMessage):
        if self == FunctionCallSupport.OPENAI_FUNCTION:
            return await self._execute_openai_function(message)
        if self == FunctionCallSupport.OPENAI_TOOL:
            return await self._execute_openai_tool(message)

    async def _execute_openai_tool(self, message: AIMessage) -> List[ToolMessage]:
        tool_calls = message.additional_kwargs.get("tool_calls", None)
        if tool_calls:
            calls = ToolCalls(calls=tool_calls)
            results = []
            for call in calls.calls:
                fname = call.function.name
                func = get_function(fname)
                if func:
                    try:
                        result = func(**call.function.arguments)
                        if iscoroutine(result):
                            result = await result
                        results.append(
                            ToolMessage(tool_call_id=call.id, content=str(result))
                        )
                    except Exception as ex:
                        logger().error(
                            f"There was an error executing tool function {fname}: {ex}"
                        )
                        results.append(
                            ToolMessage(
                                tool_call_id=call.id,
                                content=f"There was an error executing tool function {fname}: {ex}. Try to fix the error or continue without it.",
                            )
                        )
            return results

    async def _execute_openai_function(self, message: AIMessage) -> FunctionMessage:
        function_call = message.additional_kwargs.get("function_call", None)
        if function_call:
            function_call = Function(**function_call)
            fname = function_call.name
            func = get_function(fname)
            if func:
                try:
                    result = func(**function_call.arguments)
                    if iscoroutine(result):
                        result = await result
                    return FunctionMessage(content=str(result), name=fname)
                except Exception as ex:
                    logger().error(
                        f"There was an error executing function {fname}: {ex}"
                    )
                    return FunctionMessage(
                        name=fname,
                        content=f"There was an error executing tool function {fname}: {ex}. Try to fix the error or continue without it.",
                    )

    def get_definition_generator(self):
        if self == FunctionCallSupport.OPENAI_FUNCTION:
            return convert_to_openai_function
        elif self == FunctionCallSupport.OPENAI_TOOL:
            return convert_to_openai_tool

        return None

    @cache
    def get_function_definition(self, func_name: str):
        generator = self.get_definition_generator()
        func = _FUNCTIONS_BY_NAME.get(func_name, None)
        return generator(func) if func else None

    def get_kwargs(self):
        app_config = AppConfig.get_instance()
        function_names = [
            name
            for name in _FUNCTIONS_BY_NAME.keys()
            if app_config.functions and name in app_config.functions
        ]
        definitions = [self.get_function_definition(name) for name in function_names]
        if definitions:
            if self == FunctionCallSupport.OPENAI_FUNCTION:
                return {"functions": definitions, "function_call": "auto"}
            elif self == FunctionCallSupport.OPENAI_TOOL:
                return {"tools": definitions, "tool_choice": "auto"}

        return {}


_DEFAULT_FUNCTION_CALL_SUPPORT = {
    # Specify which models for which api provider support which type of function calling.
    # This can be overwritten in the app config, in case new models get deployed
    # and this code cannot be updated in time
    "btp": {
        "gpt-4,gpt-4-turbo,gpt-35-turbo,gpt-35-turbo-16k,gpt-4-32k": FunctionCallSupport.OPENAI_FUNCTION
    },
    "openai": {"*": FunctionCallSupport.OPENAI_TOOL},
}


def load_function_entry_points():
    for entry_point in pkg_resources.iter_entry_points("gptextual_function"):
        # Just loading the entry point is enough if each function
        # is decorated with @register_for_function_calling below
        entry_point.load()


def get_function(function_name: str):
    return _FUNCTIONS_BY_NAME.get(function_name, None)


def register_for_function_calling(func):
    _FUNCTIONS_BY_NAME[func.__name__] = func
    return func


def get_function_config(func):
    fname = func if isinstance(func, str) else func.__name__
    return AppConfig.get_instance().functions.get(fname, None)



from .conversation import Conversation, StreamingMessage  # noqa: F401
from .conv_manager import ConversationManager  # noqa: F401
from .models import ChatModel, ModelRegistry, AppContext  # noqa: F401



from pathlib import Path

from textual import on, log
from textual.app import ComposeResult
from textual.binding import Binding
from textual.css.query import NoMatches
from textual.events import Mount
from textual.screen import Screen
from textual.widgets import Button

from gptextual.runtime.conv_manager import ConversationManager
from gptextual.textual_ui.screens.search_screen import SearchScreen
from gptextual.textual_ui.widgets.chatbox import Chatbox
from gptextual.textual_ui.widgets.footer import CommandFooter
from gptextual.textual_ui.widgets.chat import Chat
from gptextual.textual_ui.widgets.header import ChatHeader
from gptextual.textual_ui.widgets.chat_list import ChatList
from gptextual.textual_ui.widgets.model_select import ModelSelect
from gptextual.textual_ui.screens.log_screen import LogScreen
from gptextual.textual_ui.widgets.search import Search


class ChatScreen(Screen):
    BINDINGS = [
        Binding("q", "quit", "Quit App", key_display="q"),
        Binding("ctrl+n", action="new_chat", description="New Chat", key_display="^n"),
        Binding("ctrl+f", action="search", description="Search", key_display="^f"),
        Binding(
            "ctrl+l", action="open_log", description="Show Log File", key_display="^l"
        ),
    ]

    def __init__(self):
        super().__init__()
        self.chat = Chat()
        self.log_screen = None

    def compose(self) -> ComposeResult:
        yield ChatList(id="chat-list")
        yield self.chat
        yield CommandFooter()

    async def on_mount(self, event: Mount) -> None:
        all_conversations = ConversationManager.all_conversations()
        if len(all_conversations):
            chat_list = self.query_one(ChatList)
            chat_list.current_chat_id = all_conversations[0].id
            await self.chat.load_conversation(chat_list.current_chat_id)

    @on(Chat.MessageSubmitted)
    def user_message_submitted(self, event: Chat.MessageSubmitted) -> None:
        # ChatList ordering will change, so we need to force an update...
        chat_list = self.query_one(ChatList)
        chat_list.current_chat_id = event.chat_id

    @on(Chat.AIResponseReceived)
    def on_ai_response(self, event: Chat.AIResponseReceived) -> None:
        chat_list = self.query_one(ChatList)
        chat_list.current_chat_id = event.chat_id
        # Calling this explictly in order to refresh the ordering of chats
        # in the list
        chat_list.reload_and_refresh()

    @on(ChatList.ChatOpened)
    async def on_chat_opened(self, event: ChatList.ChatOpened) -> None:
        self.chat.allow_input_submit = False
        await self.chat.load_conversation(event.chat_id)
        chat_list = self.query_one(ChatList)
        chat_list.current_chat_id = event.chat_id
        self.chat.allow_input_submit = True

    @on(ChatList.ChatDeleted)
    async def on_chat_deleted(self, event: ChatList.ChatDeleted) -> None:
        await self.action_new_chat()

    @on(Button.Pressed, "#cl-new-chat-button")
    async def on_new_chat_button(self, event: Button.Pressed):
        await self.action_new_chat()

    @on(Search.SearchResultSelected)
    async def on_search_selected(self, event: Search.SearchResultSelected):
        conv_id, message_id = event.conversation_id, event.message_id
        with self.app.batch_update():
            self.app.pop_screen()
            await self.on_chat_opened(ChatList.ChatOpened(conv_id))
            chat_box: Chatbox = self.chat.chatboxes_by_id.get(message_id, None)
            if chat_box:
                chat_box.focus(scroll_visible=True)

    def action_search(self):
        self.app.push_screen(SearchScreen(self))

    @on(ModelSelect.ModelSelected)
    def update_model(self, event: ModelSelect.ModelSelected) -> None:
        model = event.model

        try:
            conversation_header = self.query_one(ChatHeader)
        except NoMatches:
            log.error("Couldn't find ConversationHeader to update model name.")
        else:
            conversation_header.model_name = f"{model.name}@{model.api_provider}"

        self.app.app_context.model_name = model.name
        self.app.app_context.api_provider = model.api_provider

    async def action_new_chat(self) -> None:
        chat = self.query_one(Chat)
        await chat.prepare_for_new_chat()
        chat.chat_options.provider_select.focus()

    def action_open_log(self) -> None:
        self.app.push_screen(LogScreen())


class ChatScreenDark(ChatScreen):
    CSS_PATH = Path(__file__).parent.parent / "dark.tcss"


class ChatScreenLight(ChatScreen):
    CSS_PATH = Path(__file__).parent.parent / "light.tcss"



# Parts of this code uses code from https://github.com/Textualize/toolong:

# MIT License

# Copyright (c) 2024 Will McGugan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations


from pathlib import Path

from textual.app import ComposeResult
from textual.binding import Binding
from textual.lazy import Lazy
from textual.widgets import TabbedContent, TabPane
from textual.screen import Screen

from toolong.log_view import LogView
from toolong.help import HelpScreen


class LogScreen(Screen):
    BINDINGS = [
        Binding("q", action="close", description="Close"),
    ]

    CSS = """
    LogScreen {
        layers: overlay;
        & TabPane {           
            padding: 0;
        }
        & Tabs:focus Underline > .underline--bar {
            color: $accent;
        }        
        Underline > .underline--bar {
            color: $panel;
        }
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.app.start_log_watcher()

    async def on_mount(self) -> None:
        self.query("TabbedContent Tabs").set(display=len(self.query(TabPane)) > 1)
        active_pane = self.query_one(TabbedContent).active_pane
        if active_pane is not None:
            active_pane.query("LogView > LogLines").focus()

    def compose(self) -> ComposeResult:
        with TabbedContent():
            if self.app.merge and len(self.app.file_paths) > 1:
                tab_name = " + ".join(Path(path).name for path in self.app.file_paths)
                with TabPane(tab_name):
                    yield Lazy(
                        LogView(
                            self.app.file_paths,
                            self.app.watcher,
                            can_tail=False,
                        )
                    )
            else:
                for path in self.app.file_paths:
                    with TabPane(path):
                        yield Lazy(
                            LogView(
                                [path],
                                self.app.watcher,
                                can_tail=True,
                            )
                        )

    def action_help(self) -> None:
        self.app.push_screen(HelpScreen())

    def action_close(self):
        self.app.pop_screen()
        self.app.stop_log_watcher()



from __future__ import annotations

from dataclasses import dataclass
from langchain.schema import (
    BaseMessage,
    HumanMessage,
)

from shortuuid import ShortUUID

from textual import on, events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import ScrollableContainer, Vertical, VerticalScroll, Horizontal
from textual.message import Message
from textual.reactive import var
from textual.widget import Widget
from textual.widgets import TextArea, Button

from gptextual.config import AppConfig
from gptextual.runtime import (
    ConversationManager,
    Conversation,
    StreamingMessage,
    AppContext,
)
from gptextual.runtime.langchain.schema import new_message_of_type

from gptextual.textual_ui.widgets.typing_indicator import IsTyping
from gptextual.textual_ui.widgets.header import ChatHeader
from gptextual.textual_ui.widgets.model_select import ModelSelect
from gptextual.textual_ui.widgets.chatbox import Chatbox, ChatboxContainer


class ChatInputArea(TextArea):
    BINDINGS = [
        Binding(
            key="ctrl+s",
            action="focus('cl-option-list')",
            description="Focus List",
            key_display="^s",
        ),
        Binding(
            key="ctrl+f",
            action="search",
            description="Search",
            key_display="^f",
        ),
    ]

    class Submit(Message):
        def __init__(self, textarea: ChatInputArea) -> None:
            super().__init__()
            self.input_area = textarea

        @property
        def control(self):
            return self.input_area

    def __init__(self, chat: Chat, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.chat = chat

    def _on_focus(self, event: events.Focus) -> None:
        super()._on_focus(event)
        self.chat.scroll_to_latest_message()

    def action_search(self):
        self.screen.action_search()


class Chat(Widget):
    DEFAULT_CSS = """
    .multiline {
      height: 8;
    } 
    .singleline {
      height: 3;
    }
    
    
    """
    allow_input_submit = var(True)
    """Used to lock the chat input while the agent is responding."""

    def __init__(self) -> None:
        super().__init__()

        self.chat_container: ScrollableContainer | None = None
        self.chat_options: ModelSelect = ModelSelect()
        self.chat_id = None
        self.input_area = ChatInputArea(self, id="chat-input", classes="singleline")
        self.responding_indicator = IsTyping()
        self.responding_indicator.display = False
        self.multiline = False

        # needed for search
        self.chatboxes_by_id = {}
        self.uuid_gen = ShortUUID()

    @dataclass
    class FirstMessageSent(Message):
        chat_data: Conversation

    @dataclass
    class MessageSubmitted(Message):
        chat_id: str

    @dataclass
    class AIResponseReceived(Message):
        chat_id: str
        message: BaseMessage

    def compose(self) -> ComposeResult:
        yield ChatHeader()
        with Vertical(id="chat-input-container"):
            with Horizontal(id="chat-input-text-container"):
                yield self.input_area
                yield Button("Send", id="btn-submit")
            yield self.responding_indicator

        with VerticalScroll(id="chat-scroll-container") as vertical_scroll:
            self.chat_container = vertical_scroll
            vertical_scroll.can_focus = False

        yield self.chat_options

    @on(TextArea.Changed)
    def on_input_changed(self, event: TextArea.Changed):
        area = self.input_area

        def update_height(height):
            s = "singleline"
            m = "multiline"
            multiline = height > 1
            if multiline != self.multiline:
                self.multiline = multiline
                area.set_class(multiline, m)
                area.set_class(not multiline, s)
                self.refresh(layout=True)

        height = self.input_area.get_content_height(None, None, None)
        update_height(height)

    @on(ChatInputArea.Submit)
    async def user_chat_message_submitted(self, event: ChatInputArea.Submit) -> None:
        if self.allow_input_submit is True:
            user_message = event.input_area.text
            if len(user_message):
                event.input_area.clear()
                await self.chat(user_message)

    @property
    def current_conversation(self) -> Conversation:
        return (
            None
            if self.chat_id is None
            else ConversationManager.get_conversation(self.chat_id)
        )

    def scroll_to_latest_message(self):
        if self.chat_container is not None:
            self.chat_container.refresh()
            self.chat_container.scroll_end(animate=False)

    def update_header(self, *, title=None, model_name=None, api_provider=None):
        chat = self.current_conversation
        chat_header = self.query_one(ChatHeader)
        chat_header.title = title or (chat and chat.short_preview) or "New Chat..."
        chat_header.model_name = (
            (model_name and api_provider and f"{model_name}@{api_provider}")
            or (chat and f"{chat.model.name}@{chat.model.api_provider}")
            or ""
        )

    async def mount_chat_boxes(self, boxes: list[Chatbox]):
        for box in boxes:
            self.chatboxes_by_id[
                box.message.additional_kwargs.get("id", self.uuid_gen.random(20))
            ] = box
        containers = [
            ChatboxContainer(
                box, classes=("assistant-message" if box.is_ai_message else None)
            )
            for box in boxes
        ]
        await self.chat_container.mount_all(containers)

    async def load_conversation(self, chat_id: str) -> None:
        if chat_id is None:
            return

        if not self.chat_options.display and chat_id == self.chat_id:
            return

        assert self.chat_options is not None
        assert self.chat_container is not None

        # If the options display is visible, get rid of it.
        self.chat_options.display = False

        # Update the chat data
        await self.clear_chat_view()
        self.chat_id = chat_id

        chat = self.current_conversation
        model_name = chat.model.name
        chat_boxes = [
            Chatbox(model_name=model_name, message=message)
            for message in chat.displayable_messages
        ]

        stream_message = chat.streaming_message
        if stream_message:
            chat_boxes.append(
                Chatbox(model_name=model_name, message=stream_message.message)
            )
            # By binding the current stream to the new chatbox, we can resume the stream
            # even after the user navigated to another conversation in between
            self._bind_stream_to_chatbox(chatbox=chat_boxes[-1], conversation=chat)
            self.allow_input_submit = False
        else:
            self.allow_input_submit = True

        await self.mount_chat_boxes(chat_boxes)
        # await self.chat_container.mount_all(chat_boxes)
        self.chat_container.scroll_end(animate=False)
        self.update_header()

    async def prepare_for_new_chat(self) -> None:
        await self.clear_chat_view()
        self.update_header()
        self.chat_options.display = True

    async def clear_chat_view(self) -> None:
        assert self.chat_container is not None
        if self.current_conversation:
            self.current_conversation.on_stream_chunk = None
        self.chat_id = None
        self.chatboxes_by_id.clear()
        # Copy list to not modify list during loop
        children = list(self.chat_container.children)
        for child in children:
            await child.remove()

    async def chat(self, content: str) -> None:
        if not self.app.app_context.current_model:
            self.notify("No model selected", title="Error", severity="error", timeout=5)
            return

        message = new_message_of_type(HumanMessage, content=content)
        message.additional_kwargs["id"] = self.uuid_gen.random(20)
        await self.progress_conversation(message)

    async def progress_conversation(self, message: BaseMessage) -> None:
        new_chat = self.chat_id is None
        if new_chat:
            assert self.chat_options is not None
            self.chat_options.display = False
            self.chat_id = ConversationManager.create_conversation(
                model=self.app.app_context.current_model
            )
            self.update_header(
                title=Conversation.preview_from_messages([message]),
            )

        assert self.chat_id is not None
        assert message is not None

        user_message_chatbox = Chatbox(
            message=message, model_name=self.current_conversation.model.name
        )

        assert self.chat_container is not None
        await self.mount_chat_boxes([user_message_chatbox])
        # await self.chat_container.mount(user_message_chatbox)
        self.scroll_to_latest_message()

        self.responding_indicator.display = True
        self.allow_input_submit = False

        def stream_in_background():
            self.run_worker(
                self._stream_llm_response(
                    conversation=self.current_conversation, message=message
                ),
                group="llm_stream",
                exclusive=False,
            )

        self.app.call_after_refresh(stream_in_background)

    async def _stream_llm_response(
        self, *, conversation: Conversation, message: BaseMessage
    ) -> None:
        response: BaseMessage = None
        response_chatbox = None

        async def create_response_chatbox():
            response_chatbox = Chatbox(model_name=conversation.model.name)
            assert (
                self.chat_container is not None
            ), "Textual has mounted container at this point in the lifecycle."
            await self.mount_chat_boxes([response_chatbox])
            self._bind_stream_to_chatbox(
                chatbox=response_chatbox, conversation=conversation
            )
            return response_chatbox

        async for resp in conversation.progress(message):
            response = resp
            if response_chatbox is None:
                chunk = response
                if isinstance(response, StreamingMessage):
                    chunk = response.message
                response_chatbox = await create_response_chatbox()
                await conversation.on_stream_chunk(chunk, 0)
            if isinstance(response, BaseMessage):
                """
                This logic ensures that a conversation can also create more than one
                full response messages. 
                """
                if "id" in response_chatbox.message.additional_kwargs:
                    self.chatboxes_by_id[
                        response_chatbox.message.additional_kwargs["id"]
                    ] = response_chatbox
                response_chatbox = None
                conversation.on_stream_chunk = None

        self.responding_indicator = self.query_one(IsTyping)
        self.responding_indicator.display = False
        self.allow_input_submit = True
        self.post_message(
            self.AIResponseReceived(chat_id=conversation.id, message=response)
        )
        self.input_area.focus()

    def _bind_stream_to_chatbox(
        self, *, chatbox: Chatbox, conversation: Conversation
    ) -> None:
        """
        Creates a closure which is bound to the given chatbox widget.

        It is set on the conversation object as the on_stream_chunk callback.
        This way, when a user switches chats while an LLM is streaming a response,
        the stream can later be resumed by creating a new closure for a new chatbox widget.
        The new widget will then display the current stream.
        """
        config = AppConfig.get_instance().textual

        async def func(chunk, chunk_no, chatbox=chatbox):
            if isinstance(chunk, StreamingMessage):
                chatbox.message = chunk.message
                if chunk_no == 0:
                    self.post_message(self.MessageSubmitted(self.chat_id))
                if chunk_no % config.refresh_no_stream_chunks == 0:
                    scroll = True
                else:
                    scroll = False
            elif isinstance(chunk, BaseMessage):
                chatbox.message = chunk
                scroll = True

            if scroll:
                chatbox.refresh(layout=True)
                scroll_y = self.chat_container.scroll_y
                max_scroll_y = self.chat_container.max_scroll_y
                if scroll_y in range(max_scroll_y - 3, max_scroll_y + 1):
                    self.chat_container.scroll_end(animate=False)

        conversation.on_stream_chunk = func

    async def on_mount(self, _: events.Mount) -> None:
        """
        When the component is mounted, we need to check if there is a new chat to start
        """
        app_context: AppContext = self.app.app_context  # type: ignore[attr-defined]
        if app_context.chat_message is not None:
            await self.prepare_for_new_chat()
            await self.chat(app_context.chat_message)
            self.input_area.focus()

    @on(Button.Pressed, selector="#btn-submit")
    def on_submit(self, event: Button.Pressed):
        event.stop()
        self.input_area.post_message(ChatInputArea.Submit(self.input_area))



from __future__ import annotations

import re
from langchain.schema import BaseMessage, AIMessage

import pyperclip

from rich.console import RenderableType
from rich.markdown import Markdown
from textual.binding import Binding
from textual.geometry import Size
from textual.widget import Widget
from textual.containers import Container

from gptextual.textual_ui.screens.message_info_modal import MessageInfo
from gptextual.utils import format_timestamp
from gptextual.runtime.langchain.schema import new_message_of_type


class ChatboxContainer(Container):
    ...


class Chatbox(Widget, can_focus=True):
    BINDINGS = [
        Binding(
            key="ctrl+s",
            action="focus('cl-option-list')",
            description="Focus List",
            key_display="^s",
        ),
        Binding(
            key="i",
            action="focus('chat-input')",
            description="Focus Input",
            key_display="i",
        ),
        Binding(
            key="d", action="details", description="Message details", key_display="d"
        ),
        Binding(key="c", action="copy", description="Copy Message", key_display="c"),
        Binding(key="`", action="copy_code", description="Copy Code Blocks"),
    ]

    def __init__(
        self,
        *,
        model_name: str,
        message: BaseMessage | None = None,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self._message = message or new_message_of_type(AIMessage)

        self.model_name = model_name
        timestamp = format_timestamp(
            self.message.additional_kwargs.get("timestamp", 0) or 0
        )
        self.tooltip = f"Sent {timestamp}"

    @property
    def is_ai_message(self):
        return self.message and self.message.type.lower().startswith("ai")

    def on_mount(self) -> None:
        if self.message.type.lower().startswith("ai"):
            self.add_class("assistant-message")

    def get_code_blocks(self, markdown_string):
        pattern = r"```(.*?)\n(.*?)```"
        code_blocks = re.findall(pattern, markdown_string, re.DOTALL)
        return code_blocks

    def action_copy_code(self):
        codeblocks = self.get_code_blocks(self.message.content)
        output = ""
        if codeblocks:
            for lang, code in codeblocks:
                output += f"{code}\n\n"
            pyperclip.copy(output)
            self.notify("Codeblocks have been copied to clipboard", timeout=3)
        else:
            self.notify("There are no codeblocks in the message to copy", timeout=3)

    def action_copy(self) -> None:
        pyperclip.copy(self.message.content)
        self.notify("Message content has been copied to clipboard", timeout=3)

    def action_details(self) -> None:
        self.app.push_screen(
            MessageInfo(message=self.message, model_name=self.model_name)
        )

    @property
    def markdown(self) -> Markdown:
        return Markdown(self.message.content or "")

    def render(self) -> RenderableType:
        return self.markdown

    def get_content_width(self, container: Size, viewport: Size) -> int:
        # Naive approach. Can sometimes look strange, but works well enough.
        content = self.message.content or ""
        return min(len(content), container.width)

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, m):
        if not isinstance(m, BaseMessage):
            raise ValueError("Message must be a BaseMessage instance")
        self._message = m



from setuptools import setup, find_packages

setup(
    name="gptextual",
    version="0.0.3",
    packages=find_packages(exclude=["tests"]),
    package_data={"": ["*.tcss"]},
    include_package_data=True,
    description="A configurable, terminal based Chat client for LLMs built with Textual",
    author="Stefan Kirchfeld",
    author_email="stefan.kirchfeld@gmail.com",
    url="https://github.com/stefankirchfeld/gptextual",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    entry_points={
        "console_scripts": [
            "gptx = gptextual.textual_ui.app:run",
        ],
        "gptextual_function": [
            "google_web_search = gptextual.runtime.function_calling.functions:google_web_search"
        ],
    },
    keywords="gpt,chat,llm,chatgpt,langchain,openai,sap,textual,terminal",
    install_requires=[
        "httpx~=0.26.0",
        "humanize~=4.9.0",
        "langchain~=0.1.5",
        "langchain-core~=0.1.19",
        "polars~=0.20.7",
        "pydantic~=2.6.1",
        "pyperclip~=1.8.2",
        "rich~=13.7.0",
        "shortuuid~=1.0.11",
        "textual==0.50.1",
        "tiktoken~=0.5.2",
        "toolong==1.2.0",
        "pyyaml~=6.0.1",
        "setuptools~=69.1.0",
    ],
    extras_require={
        "openai": ["langchain-openai~=0.0.5"],
        "google": ["langchain-google-genai~=0.0.9"],
        "sap": ["generative-ai-hub-sdk~=1.2.2"],
        "all": [
            "langchain-openai~=0.0.5",
            "langchain-google-genai~=0.0.9",
            "generative-ai-hub-sdk~=1.2.2",
        ],
    },
)



from __future__ import annotations

import httpx

from .function_call_support import (
    register_for_function_calling,
    get_function_config,
)


def _process_search_results(result_json):
    # Initialize an empty list to hold the results
    results = []

    # Get the items from the result JSON
    items = result_json.get("items", [])

    # Loop through the top 3 items
    for i, item in enumerate(items[:10]):
        # Prepare the result dictionary
        result = f"Title: {item.get('title')}|URL: {item.get('link')}|Snippet: {item.get('snippet')}"

        # Append the result to the list
        results.append(result)

    return "\n".join(results)


@register_for_function_calling
async def google_web_search(query: str) -> str:
    """
    Executes a Google search with the specified search string and returns the top 10 search results.
    For each result, a title, URL and preview snippet is returned.

    Args:
        query: the query string
    """
    config = get_function_config(google_web_search)
    api_key, cx_id = None, None
    if config:
        api_key = config.get("api_key", None)
        cx_id = config.get("cx_id", None)

    if api_key is None or cx_id is None:
        return "This tool is not configured correctly and cannot be used at this time. Please try to continue without it or let the user know."

    url = "https://www.googleapis.com/customsearch/v1"

    params = {"key": api_key, "cx": cx_id, "q": query}

    resp = None
    async with httpx.AsyncClient() as client:
        resp = await client.get(url, params=params, timeout=5)

    resp.raise_for_status()  # Raises an exception if the HTTP status is 400 or higher
    result_json = resp.json()  # Get the response body as JSON

    return _process_search_results(result_json)



from __future__ import annotations

import glob
import os
from pathlib import Path
from dataclasses import dataclass
import threading

from gptextual.logging import logger

from .models import ChatModel
from .conversation import Conversation, conversation_path


@dataclass
class ConversationManager:
    conversations = {}
    # Needed so we can monkey patch in an extension package
    conversation_class = Conversation

    @classmethod
    def load_conversations_from_storage(cls):
        # Get a list of all JSON files in the storage path
        files = glob.glob(str(conversation_path / "*.json"))

        if len(files):
            cls.conversations = {
                key: val
                for key, val in (
                    (Path(file).stem, cls.conversation_class.load(Path(file).stem))
                    for file in files
                )
                if val is not None
            }

    @classmethod
    def save_all(cls):
        def do_save():
            for conversation in ConversationManager.all_conversations():
                try:
                    conversation.save(in_background=False)
                except Exception as ex:
                    logger().error(f"Error saving conversations, {ex}")

        save_thread = threading.Thread(target=do_save)
        save_thread.start()

    @classmethod
    def delete_conversation(cls, id: str):
        header_file = conversation_path / f"{id}.json"
        messages_file = conversation_path / f"{id}.parquet"
        try:
            os.remove(header_file)
            os.remove(messages_file)
        except FileNotFoundError:
            pass
        except Exception as e:
            logger().error(f"Error deleting conversation with id {id}: {e}")
        if id in cls.conversations:
            del cls.conversations[id]

    @classmethod
    def all_conversations(cls):
        return list(
            sorted(
                [data for data in cls.conversations.values()],
                key=lambda x: x.update_time,
                reverse=True,
            )
        )

    @classmethod
    def get_conversation(cls, conversation_id: str):
        return cls.conversations.get(conversation_id, None)

    @classmethod
    def create_conversation(cls, *, model: ChatModel, system_message=None) -> str:
        conv = cls.conversation_class.create_new(
            model=model,
            system_message=system_message,
        )
        cls.conversations[conv.id] = conv
        return conv.id
